from loguru import logger
import os
import pandas as pd
from pandas.core import base
from haferml.blend.config import construct_paths
import joblib
import simplejson as json
from haferml.sync.local import isoencode
from haferml.model.dataset import DataSet
from haferml.model.modelset import ModelSet
from haferml.model.workflow import ModelWorkflow


class DataSetX(DataSet):
    """
    DataSetX class deals with the dataset to be used in a model


    :param config: a dictionary that contains the configurations.
    :type config: dict
    :param base_folder: working directory where all the artifacts are being perserved.
    :type base_folder: str
    """

    def __init__(self, config, base_folder):
        super(DataSetX, self).__init__(config, base_folder)

        self.targets = self.config.get("targets")
        self.features = self.config.get("features")
        self.artifacts = self.config["artifacts"]

        logger.debug(f"features: {self.features}\n" f"predict: {self.targets}")

    def create_train_test_datasets(self, dataframe):
        """
        create_train_test_datasets will create

        - self.data: the full input data right before train test split
        - self.X_train, self.y_train, self.X_test, self.y_test

        :param dataframe: the dataframe to be splitted.
        :type dataframe: pandas.DataFrame
        """

        raise Exception("create_train_test_dataset has not yet been implemented!")

    @staticmethod
    def _save_data(dataframe, destination):
        """
        `_save_data` saves the dataframe locally.

        :param dataframe: dataframe to be saved
        :type dataframe: pandas.DataFrame
        :param destination: where the data is saved
        :type destination: str
        """
        logger.info("Export test and train data")
        dataframe.to_parquet(destination)

    def _export_train_test_data(self):
        """
        _export_train_test_data saves train and test datasets
        """

        dataset_folder = construct_paths(
            self.artifacts["dataset"], base_folder=self.base_folder
        )["local"]

        ## Save the train and test datasets
        logger.info("Export test and train data")
        self._save_data(
            self.X_train, os.path.join(dataset_folder, "model_X_train.parquet")
        )

        self._save_data(
            self.X_test, os.path.join(dataset_folder, "model_X_test.parquet")
        )

        self._save_data(
            pd.DataFrame(self.y_train, columns=self.pred_cols),
            os.path.join(dataset_folder, "model_y_train.parquet"),
        )

        self._save_data(
            pd.DataFrame(self.y_test, columns=self.pred_cols),
            os.path.join(dataset_folder, "model_y_test.parquet"),
        )

        # Save dataset locally
        self._save_data(self.data, os.path.join(dataset_folder, "dataset.parquet"))


class ModelSetX(ModelSet):
    """
    The core of the model including hyperparameters.

    :param config: a dictionary that contains the configurations.
    :type config: dict
    :param base_folder: working directory where all the artifacts are being perserved.
    :type base_folder: str
    """

    def __init__(self, config, base_folder):
        super(ModelSet, self).__init__(config, base_folder)

        self.targets = self.config.get("targets")
        self.features = self.config.get("features")
        self.artifacts = self.config["artifacts"]

        logger.debug(f"features: {self.features}\n" f"predict: {self.targets}")

    def create_model(self):
        ...

    @property
    def hyperparameters(self):
        return self._set_hyperparameters()

    def _set_hyperparameters(self):
        """
        _set_hyperparameters creates hyperpamater grid
        """
        ...


class ModelWorkflowX(ModelWorkflow):
    """ModelWorkflowX class that holds DataSetX and ModelSetX

    :param config: a dictionary that contains the configs.
    :type config: dict
    :param dataset: a DataSet object that contains the data and provides a `create_train_test_datasets` method.
    :type dataset: haferml.model.DataSet
    :param modelset: a ModelSet object that contains the model as well as the hyperparameters and a `create_model`.
    :type modelset: haferml.model.ModelSet
    :param base_folder: working directory where all the artifacts are being perserved.
    :type base_folder: str
    """

    def __init__(self, config, dataset, modelset, base_folder):
        self.config = config
        self.base_folder = base_folder
        self.DataSet = dataset
        self.ModelSet = modelset
        self.artifacts = self.config["artifacts"]

    def fit_and_report(self):
        """
        _fit_and_report fits the model using input data and generate reports
        """

        logger.info("Fitting the model ...")
        logger.debug(
            "Shape of train data:\n"
            f"X_train: {self.DataSet.X_train.shape}, {self.DataSet.X_train.sample(3)}\n"
            f"y_train: {self.DataSet.y_train.shape}, {self.DataSet.y_train.sample(3)}"
        )
        self.ModelSet.model.fit(
            self.DataSet.X_train.squeeze(), self.DataSet.y_train.squeeze()
        )

        self.report = {
            "hyperparameters": self.ModelSet.hyperparams_grid,
            "best_params": self.ModelSet.model.best_params_,
            "cv_results": self.ModelSet.model.cv_results_,
        }

        logger.debug(self.report)

    def export_results(self):
        """
        export_results saves the necessary artifacts
        """
        model_artifacts = construct_paths(
            self.artifacts["model"], base_folder=self.base_folder
        )
        model_folder = model_artifacts["local"]
        model_path = model_artifacts["file_path"]

        if not os.path.exists(model_folder):
            os.makedirs(model_folder)

        logger.info("Preserving models ...")
        joblib.dump(self.ModelSet.model, model_path)

        logger.info("Perserving logs ...")
        log_file_path = f"{model_path}.log"
        logger.info(f"Save log file to {log_file_path}")
        with open(log_file_path, "a+") as fp:
            json.dump(self.report, fp, default=isoencode)
            fp.write("\n")
        logger.info(f"Saved logs")

    def train(self, dataset):
        """
        train connects the training workflow

        :param dataset: dataframe being used to train the model
        :type dataset: pandas.DataFrame
        """

        logger.info("1. Create train test datasets")
        self.DataSet.create_train_test_datasets(dataset)
        logger.info("2. Create model")
        self.ModelSet.create_model()
        logger.info("3. Fit model and report")
        self.fit_and_report()
        logger.info("4. Export results")
        self.export_results()



from loguru import logger


class DataSet:
    """
    DataSet to be used in a model workflow.

    !!! warning
        This class is a proposed framework. There are many member functions to be implemented if you are using the `DataSet` class.

        In most projects, we do not need to use this class. We have a class `DataSetX` which is a more detailed implementation in our `model.pipeline` module.


    The interface of this class in a workflow is the `create_train_test_datasets` method.

    Depending on the requirements, one can also implement a method `_export_train_test_data` to save the dataset locally or remotely.

    """

    def __init__(self, config, base_folder):
        self.config = config
        self.base_folder = base_folder

    def create_train_test_datasets(self):
        """
        create_train_test_datasets is our interface to our workflow.

        !!! warning
            Please implement this method.

        One could call the `_export_train_test_data` method to export the dataset locally or remotely.
        """
        ...

    @staticmethod
    def _save_data(dataframe, destination):
        """
        `_save_data` is dummy method method to save the data to a local destination.

        !!! warning
            Please implement this method.

        :param dataframe: a pandas dataframe that holds the data
        :type dataframe: pandas.DataFrame
        :param destination: destination of the data on a local machine
        :type destination: str
        """
        logger.debug(f"Saving data to {destination} ...")
        dataframe.to_parquet(destination)

    def _export_train_test_data(self):
        """
        _export_train_test_data saves train and test datasets.

        !!! warning
            Please implement this method.

        """
        ...



from loguru import logger


class ModelSet:
    """
    `ModelSet` is the core of the model including hyperparameters.

    !!! warning
        This class is a proposed framework. There are many member functions to be implemented if you are using the `ModelSet` class.

        In most projects, we do not need to use this class. We have a class `ModelSetX` which is a more detailed implementation in our `model.pipeline` module.

    """

    def __init__(self, config, base_folder):
        self.config = config
        self.base_folder = base_folder

        logger.debug(
            f"base folder: {self.base_folder}\n" f"artifacts configs: {self.artifacts}"
        )

    def create_model(self):
        """
        `create_model` creates the model and updates the property `self.model`.
        """
        ...

    @property
    def hyperparameters(self):
        """
        `hyperparameters` specifies the hyperparameters. This is a property.
        """
        ...



import pandas as pd
from haferml.preprocess.ingredients import OrderedProcessor, attributes
from loguru import logger


class BasePreProcessor(OrderedProcessor):
    """
    Shared methods to transform the datasets

    The following example demonstrates how to use it.

    ```python
    from haferml.data.preprocessing.ingredients import OrderedProcessor, attributes

    class DemoPreProcessor(BasePreProcessor):
        def __init__(self, config, columns, cutoff_timestamp=None):
            super(DemoPreProcessor, self).__init__(
                config=config, columns=columns
            )

            self.cutoff_timestamp=cutoff_timestamp

        def merge_datasets(self, datasets):

            df_a = datasets["a"]
            df_b = datasets["b"]

            # 1. We only take data that is later than a certain date

            if self.cutoff_timestamp:
                filter_a_mask = (
                    df_a.req_created_at > self.cutoff_timestamp
                )

            df_a = df_a.loc[filter_a_mask]

            # combine dataset a and b
            dataset = pd.merge(
                df_a,
                df_b,
                how="left",
                on="request_id",
            )

            return dataset

        @attributes(order=1)
        def _fix_names(self, dataset):

            dataset["names"] = dataset.names.replace(
                "Tima", "Tim"
            )

        @attributes(order=2)
        def _convert_requirement_to_bool(self, dataset):

            dataset[
                "requirements"
            ] = dataset.requirements.apply(
                lambda x: False if pd.isnull(x) else True
            )

        @attributes(order=12)
        def _filter_columns_and_crossing(self, dataset):
            # _filter_columns_and_crossing removes unnecessary columns and append crossings

            # only keep the specified columns
            # the following code also deals with feature crossings
            if self.columns:
                columns = list(set(self.columns))
                crossing = []
                features_ex = []
                for col in columns:
                    if "__" in col:
                        crossing.append(col)
                    else:
                        features_ex.append(col)

                dataset = dataset[features_ex]

                for fc in crossing:
                    fc_cols = fc.split("__")
                    fc_series = dataset[fc_cols[0]]
                    for fc_col in fc_cols[1:]:
                        fc_series = fc_series * dataset[fc_col]
                    dataset[fc] = fc_series


    dp = DemoPreProcessor(config={}, columns=['names', 'requirements', 'names__requirements'])
    dataset = {
        "a": pd.DataFrame([{"names": "Tima Cook", "requirements": "I need it"}]),
        "b": pd.DataFrame([{"names": "Time Cook", "requirements": None}])
    }
    dp.preprocess(dataset)
    ```
    """

    def __init__(self, config, columns):
        super(BasePreProcessor, self).__init__(config=config, columns=columns)

        self.config = self.params.get("config")
        logger.info(f"Applying config:\n{self.config}")
        self.columns = self.params.get("columns")
        logger.info(f"Applying columns:\n{self.columns}")

    def merge_datasets(self, datasets):
        """
        merge_datasets merges the datasets into one singular dataframe to be processed.

        :param datasets: dictionary that contains the dataframes
        :type datasets: dict
        """

        raise NotImplementedError("Please implement this method!")

    def preprocess(self, datasets, **params):
        """
        preprocess connects the transforms into pipelines

        :param datasets: input datasets as list or dict of single dataframe
        :param merge: whether and how to merge datasets: True or False
        """

        if params.get("merge") is True:
            dataframe = self.merge_datasets(datasets)
        elif isinstance(datasets, dict):
            logger.warning(
                "No specific merge methods specified\n"
                "Auto concating all the datasets"
            )
            dataframe = pd.concat(datasets.values())
        elif isinstance(datasets, list):
            logger.warning(
                "No specific merge methods specified\n"
                "Auto concating all the datasets"
            )
            dataframe = pd.concat(datasets)
        elif isinstance(datasets, pd.DataFrame):
            logger.info(
                "Input dataset is a single dataframe, " "making a copy for safety"
            )
            dataframe = datasets.copy()
        else:
            raise TypeError("Input datasets should be dataframe or list of dataframes")

        # Go through the transforms
        for t in self.transforms:
            logger.info(f"Performing {t} ...")
            self.transforms[t](dataframe)
            logger.info(f"{t} is done.")

        return dataframe


if __name__ == "__main__":

    from haferml.preprocess.ingredients import attributes

    class DemoPreProcessor(BasePreProcessor):
        def __init__(self, config, columns, cutoff_timestamp=None):
            super(DemoPreProcessor, self).__init__(config=config, columns=columns)

            self.cutoff_timestamp = cutoff_timestamp

        def merge_datasets(self, datasets):
            """
            merge_datasets merges the dataframes in the datasets
            """
            df_a = datasets["a"]
            df_b = datasets["b"]

            # 1. We only take data that is later than a certain date

            if self.cutoff_timestamp:
                filter_a_mask = df_a.req_created_at > self.cutoff_timestamp

            df_a = df_a.loc[filter_a_mask]

            # combine dataset a and b
            dataset = pd.merge(
                df_a,
                df_b,
                how="left",
                on="request_id",
            )

            return dataset

        @attributes(order=1)
        def _fix_names(self, dataset):

            dataset["names"] = dataset.names.replace("Tima", "Tim")

        @attributes(order=2)
        def _convert_requirement_to_bool(self, dataset):

            dataset["requirements"] = dataset.requirements.apply(
                lambda x: False if pd.isnull(x) else True
            )

        @attributes(order=12)
        def _filter_columns_and_crossing(self, dataset):
            """
            _filter_columns_and_crossing removes unnecessary columns and append crossings
            """

            # only keep the specified columns
            # the following code also deals with feature crossings
            if self.columns:
                columns = list(set(self.columns))
                crossing = []
                features_ex = []
                for col in columns:
                    if "__" in col:
                        crossing.append(col)
                    else:
                        features_ex.append(col)

                dataset = dataset[features_ex]

                for fc in crossing:
                    fc_cols = fc.split("__")
                    fc_series = dataset[fc_cols[0]]
                    for fc_col in fc_cols[1:]:
                        fc_series = fc_series * dataset[fc_col]
                    dataset[fc] = fc_series

    dp = DemoPreProcessor(
        config={}, columns=["names", "requirements", "names__requirements"]
    )
    dataset = {
        "a": pd.DataFrame([{"names": "Tima Cook", "requirements": "I need it"}]),
        "b": pd.DataFrame([{"names": "Time Cook", "requirements": None}]),
    }
    dp.preprocess(dataset)



import datetime
from loguru import logger

import dateutil
import pandas as pd


def convert_to_datetime(input_date, dayfirst=None, input_tz=None, output_tz=None):
    """
    Convert input to *datetime* object.
    This is the last effort of converting input to datetime.
    The order of instance check is
    1. datetime.datetime
    2. str
    3. float or int

    ```
    >>> handle_strange_dates(1531323212311)
    datetime(2018, 7, 11, 17, 33, 32, 311000)
    >>> handle_strange_dates(datetime(2085,1,1))
    datetime(2050, 1, 1)
    ```

    :param input_date: input data of any possible format
    :param input_tz: input timezone, defaults to utc
    :param output_tz: output timezone, defaults to utc
    :return: converted datetime format
    :rtype: datetime.datetime
    """
    if dayfirst is None:
        dayfirst = True
    if input_tz is None:
        input_tz = datetime.timezone.utc
    if output_tz is None:
        output_tz = datetime.timezone.utc

    res = None
    if isinstance(input_date, datetime.datetime):
        res = input_date
        if input_tz:
            res = res.replace(tzinfo=input_tz)
        if output_tz:
            res = res.astimezone(output_tz)
    elif isinstance(input_date, str):
        try:
            res = dateutil.parser.parse(input_date, dayfirst=dayfirst)
            if input_tz:
                res = res.replace(tzinfo=input_tz)
            if output_tz:
                res = res.astimezone(output_tz)
        except:
            logger.warning(f"Could not convert {input_date} to datetime!")
            pass
    elif isinstance(input_date, (float, int)):
        try:
            res = datetime.datetime.utcfromtimestamp(input_date / 1000)
            if input_tz:
                res = res.replace(tzinfo=input_tz)
            if output_tz:
                res = res.astimezone(output_tz)
        except:
            logger.warning(f"Could not convert {input_date} to datetime!")
            pass
    else:
        raise Exception(
            "Could not convert {} to datetime: type {} is not handled".format(
                input_date, type(input_date)
            )
        )

    return res


def unpack_datetime(data):
    """
    unpack_datetime converts datetime (string) to a dict of useful date information
    """
    res = {}
    dt = convert_to_datetime(data, dayfirst=False)
    if dt:
        try:
            res["year"] = dt.year
        except Exception as e:
            logger.error(f"Could not find year for {dt} (raw: {data})")

        try:
            res["month"] = dt.month
        except Exception as e:
            logger.error(f"Could not find month for {dt} (raw: {data})")

        try:
            res["day"] = dt.day
        except Exception as e:
            logger.error(f"Could not find day for {dt} (raw: {data})")

        try:
            res["weekday"] = dt.weekday() + 1
        except Exception as e:
            logger.error(f"Could not find weekday for {dt} (raw: {data})")

    return res


def date_range_has_weekday(dt_start, dt_end):
    """
    date_range_has_weekday decides if the given date range contains weekday

    :param dt_start: datetime of the start of date range
    :param dt_end: datetime of the end of date range
    """
    res = []

    if pd.isnull(dt_start) or pd.isnull(dt_end):
        logger.warning(f"date start end not specified: {dt_start}, {dt_end}")
        return None

    if isinstance(dt_start, str):
        dt_start = pd.to_datetime(dt_start)
    if isinstance(dt_end, str):
        dt_end = pd.to_datetime(dt_end)

    for dt in pd.date_range(dt_start, dt_end):
        if dt.weekday() < 5:
            res.append(True)
        else:
            res.append(False)

    return True in res



import os
from awscli.clidriver import create_clidriver
import boto3


def aws_cli(*cmd):
    """
    aws_cli invokes the aws cli processes in python to execute awscli commands.

    !!! warning
        This is not the most elegant way of using awscli.
        However, it has been a convinient function in data science projects.

        This function is adapted from https://github.com/boto/boto3/issues/358#issuecomment-372086466


    AWS credential env variables should be configured before calling this function.
    The awscli command should be wrapped as a tuple. To download data from S3 to a local path, use

    ```
    >>> aws_cli(('s3', 'sync', 's3://s2-fpd/augmentation/', '/tmp/test'))
    Similarly, upload is done in the following way
    >>> # local_path = ''
    >>> # remote_path = ''
    >>> _aws_cli(('s3', 'sync', local_path, remote_path))
    ```

    :param *cmd: tuple of awscli command.
    """
    old_env = dict(os.environ)
    try:
        # Set up environment
        env = os.environ.copy()
        env["LC_CTYPE"] = "en_US.UTF"
        os.environ.update(env)

        # Run awscli in the same process
        exit_code = create_clidriver().main(*cmd)

        # Deal with problems
        if exit_code > 0:
            raise RuntimeError(f"AWS CLI exited with code {exit_code}")
    finally:
        os.environ.clear()
        os.environ.update(old_env)


def s3_download(path, folder):
    """
    `s3_download` downloads files from S3.

    ```
    >>> s3_download(config_path, base_folder)
    ```

    :param path: s3 uri
    :type path: str
    :param folder: destination folder
    :type folder: str
    """

    if not path.startswith("s3://"):
        raise Exception(f"{path} is not S3 uri!")
    else:
        # e.g., s3://mein-work/abc/performance/model_performance_log.json
        s3_bucket = path.split("/")[2]
        s3_filepath = "/".join(path.split("/")[3:])
        # get the name of the config file
        s3_filename = path.split("/")[-1]
        # local config path is constructed from base folder and filename
        path = os.path.join(folder, s3_filename)
        s3 = boto3.client("s3")
        s3.download_file(s3_bucket, s3_filepath, path)



import inspect
import re
import time

import haferml.data.wrangle as wlg
import pandas as pd
from loguru import logger


class Transformer:
    """
    Base Transformer of a single data record
    """

    def __init__(self, schema, use_schema_column=None):

        if use_schema_column is None:
            use_schema_column = "source_column"
        self.use_schema_column = use_schema_column
        # specify the schema content to use
        self.schema = schema
        self._schema_to_utils()

        # get current timestamp
        self.current_timestamp = int(time.time())

    def _schema_to_utils(self):
        """
        _schema_to_utils converts some utility schemas using the input full schema
        """
        self.column_rename_schema = {
            i.get("column_name"): i.get("column_name") for i in self.schema
        }

        # build transformer schema
        self.transformer_schema = {
            i.get("column_name"): {"type": i.get("type")} for i in self.schema
        }
        self._get_transformers()  # enhances the schema with the transformer function

    def _get_transformers(self):
        """
        _get_transformers extracts the list of transformers
        """
        re_transformer_name = re.compile("^_transformer__(.*?)$")

        transformers = {}
        all_methods = dict(inspect.getmembers(self))
        for i in all_methods:
            if i.startswith("_transformer__"):
                transformer_name = re_transformer_name.findall(i)[0]
                transformer_method_i = all_methods.get(i)
                transformers[transformer_name] = transformer_method_i

        logger.debug("All methods: {}".format(all_methods))
        logger.info("All predefined transformers: {}".format(transformers))

        for i in self.transformer_schema:
            logger.info("this transformer schema: {}".format(i))
            i_val = self.transformer_schema.get(i, {})
            if i in transformers:
                i_val["transformer"] = transformers.get(i)
                logger.info("Has predefined transformer for {}".format(i))
            else:
                logger.info(
                    "Using default transformer for {}; format: {}".format(
                        i, i_val.get("type")
                    )
                )
                to_format_type = i_val.get("type")
                i_val["transformer"] = self._universal_transformer(
                    to_format=to_format_type
                )

            self.transformer_schema[i] = i_val

    @staticmethod
    def _universal_transformer(to_format, from_format=None):
        """
        _general_transformer is to be used if a specific transformer of the column is not found
        """

        def transformer(data):
            """
            transformer is the actual transformer
            """
            if pd.isnull(data):
                return None

            logger.debug(f"Transforming {data} to format {to_format}")
            if to_format.lower() in ("str", "string"):
                try:
                    res = str(data)
                    res = res.strip()
                except Exception as e:
                    raise Exception("Could not convert {} to str".format(data))
            elif to_format.lower() == "int":
                if isinstance(data, str):
                    if ("." in data) or ("," in data):
                        data = wlg.misc.eu_float_string_to_float(data)
                try:
                    res = int(float(data))
                except Exception as e:
                    raise Exception("Could not convert {} to float->int".format(data))
            elif to_format.lower() == "float":
                if isinstance(data, str):
                    if ("." in data) or ("," in data):
                        data = wlg.misc.eu_float_string_to_float(data)
                try:
                    res = float(data)
                except Exception as e:
                    raise Exception("Could not convert {} to float".format(data))
            elif to_format.lower() == "datetime":
                res = wlg.datetime.convert_to_datetime(data, dayfirst=False)
            elif to_format.lower() == "date":
                res = wlg.datetime.convert_to_date(data)
            elif to_format.lower() == "bool":
                res = wlg.misc.convert_to_bool(data)
            elif to_format.lower() == "list":
                res = wlg.misc.convert_str_repr_to_list(data)
            else:
                raise Exception(
                    f"Can not transform {data}; No transformer defined for the format: {to_format}"
                )

            return res

        return transformer

    def transform(self, record):
        """
        transform transforms the json (list of dict data) into standardized format
        """

        # sometime the transformations requires other fields
        # we need to set self.record to access all the fields
        self.record = record.copy()
        try:
            for key, val in record.items():
                try:
                    val = self.transformer_schema[key]["transformer"](val)
                    record[key] = val
                except Exception as e:
                    logger.error(
                        "Failed to transform key, val: {}, {}; schema is {}; e {}".format(
                            key, val, self.transformer_schema[key], e
                        )
                    )
        except Exception as e:
            raise Exception(
                "Failed to transform record: {};error is {}".format(record, e)
            )

        return record



import os


def check_env(var_list):
    """
    check_env checks if the given environment variables exists
    :param var_list: list of variables to be checked
    :type definition: list
    """

    missing = []
    res = {}
    for envvar in var_list:
        envvar_val = os.getenv(envvar)
        if envvar_val is None:
            missing.append(envvar)
        res[envvar] = envvar_val

    if missing:
        raise Exception(f"Missing envs {missing}")

    return res



import os
from loguru import logger
import simplejson as json


def load_config(config_path, base_folder=None):
    """
    load_config loads the config files of the project

    :param config_path: path to the config file
    :type config_path: str, optional
    """
    if config_path is None:
        raise Exception(f"config_path has not been specified...")

    if not os.path.exists(config_path):
        raise Exception(
            f"config file path {config_path} does not exist! Beware of the relative path."
        )

    logger.debug(f"Loading config from {config_path}")
    with open(config_path, "r") as fp:
        config = json.load(fp)

    if not config:
        logger.warning(f"The config is empty: {config}")

    return config


def get_config(configs, path):
    """
    Get value of the configs under specified path

    :param dict configs: input dictionary
    :param list path: path to the value to be obtained

    ```
    >>> get_config({'etl':{'raw':{'local':'data/raw', 'remote': 's3://haferml-tutorials/rideindego/marshall/data/raw'}}},['etl','raw'])
    {'local':'data/raw', 'remote': 's3://haferml-tutorials/rideindego/marshall/data/raw'}
    ```
    """

    # Construct the path
    if not isinstance(path, (list, tuple)):
        logger.warning(f"path is not list nor tuple, converting to list: {path}")
        path = [path]

    # Find the values
    res = configs.copy()
    for p in path:
        res = res[p]

    return res


def construct_paths(config, base_folder):
    """
    construct_paths reconstructs the path based on base folder

    The `local` key will be replaced with the
    """

    if not config.get("local"):
        logger.warning(f"{config} does not contain local key ")
        return config

    config_recon = {}
    config_local = config["local"]
    config_name = config.get("name")
    config_local = os.path.join(base_folder, config_local)
    config_recon["local"] = config_local
    if config_name:
        config_local_full = os.path.join(config_local, config_name)
        config_recon["file_path"] = config_local_full

    return {**config, **config_recon}



from ast import literal_eval

from loguru import logger


def get_value_in_dict_recursively(dictionary, path, ignore_path_fail=None):
    """
    Get value of a dictionary according to specified path (names)

    :param dict dictionary: input dictionary
    :param list path: path to the value to be obtained

    This function always returns the value or None.

    ```
    >>> get_value_in_dict_recursively({'lvl_1':{'lvl_2':{'lvl_3':'lvl_3_value'}}},['lvl_1','lvl_3'])
    {'lvl_3':'lvl_3_value'}
    >>> get_value_in_dict_recursively({1:{2:{3:'hi'}}},[1,'2',3])
    {'hi'}
    ```
    """
    if ignore_path_fail is None:
        ignore_path_fail = True

    if isinstance(path, list):
        path_temp = path.copy()
    elif isinstance(path, tuple):
        path_temp = list(path).copy()
    else:
        logger.warning(f"path is not list or tuple, converting to list: {path}")
        path_temp = [path].copy()

    if len(path_temp) > 1:
        pop = path_temp.pop(0)
        try:
            pop = int(pop)
        except ValueError:
            if ignore_path_fail:
                logger.warning(f"can not get path")
                pass
            else:
                raise Exception(f"specified path ({path}) is not acceptable")

        try:
            return get_value_in_dict_recursively(dictionary[pop], path_temp)
        except:
            logger.debug(f"did not get values for {pop}")
            return None
    elif len(path_temp) == 0:
        return None
    else:
        try:
            val = int(path_temp[0])
        except:
            val = path_temp[0]
        try:
            return dictionary[val]
        except KeyError:
            logger.error(f"KeyError: Could not find {path_temp[0]}")
            return None
        except TypeError:
            logger.error(f"TypeError: Could not find {path_temp[0]}")
            return None


###############
# Generic
###############


def convert_str_repr_to_list(inp):
    """
    convert_str_repr_to_list concerts string representation of list to list
    """

    res = []
    if isinstance(inp, str):
        try:
            res = literal_eval(inp)
        except Exception as e:
            raise Exception(f"Could not convert {inp} to list")
    elif isinstance(inp, (list, tuple, set)):
        res = list(inp)

    return res


def convert_str_repr_to_tuple(inp):
    """
    convert_str_repr_to_tuple converts string representation of tuple to tuple
    """

    res = []
    if isinstance(inp, str):
        try:
            res = literal_eval(inp)
        except Exception as e:
            raise Exception(f"Could not convert {inp} to list")
    if isinstance(inp, (list, tuple, set)):
        res = tuple(inp)

    return res


def convert_to_bool(data):
    """
    convert_to_bool converts input to bool type in python.

    The following values are converted to True:

    1. 'true'
    2. 'yes'
    3. '1'
    4. 'y'
    5. 1

    The following values are converted to False:

    1. 'false'
    2. 'no'
    3. '0'
    4. 'n'
    5. 0

    :param data: input data
    :return: boolean value of the input data
    :rtype: bool
    """
    res = None
    if data is None:
        return res
    elif isinstance(data, bool):
        res = data
    elif isinstance(data, str):
        if data.lower().strip() in ["true", "yes", "1", "y"]:
            res = True
        elif data.lower().strip() in ["false", "no", "0", "n"]:
            res = False
        else:
            res = None
    elif isinstance(data, (float, int)):
        res = bool(data)

    return res


def eu_float_string_to_float(data):
    """
    eu_float_string_to_float converts strings in EU format to floats

    :param data: string of the float in EU conventions
    :type data: str
    :return: converted float from the string
    :rtype: float
    """
    if isinstance(data, str):
        res = data.replace(".", "")
        res = res.replace(",", ".")
        try:
            res = float(res)
        except Exception as e:
            raise Exception(f"Could not convert string {data} to float: {e}")
    else:
        raise TypeError("Input data should be string")

    return res



import json
import os
import datetime
import numpy as np

from loguru import logger


def prepare_folders(folder_list=None, base_folder=None):
    """
    prepare_folders creates the necessary folders

    :param base_folder: base folder of the whole project
    :type base_folder: str
    :param folder_list: list of folders to create, relative to base_folder
    :type folder_list: list
    """
    if folder_list is None:
        raise Exception("Please specify the list of folder using fodler_list")

    if os.path.exists(base_folder):
        logger.info(f"Using base folder {base_folder}!")

    # prepare the model folder
    if isinstance(folder_list, (tuple, list, set)):
        pass
    elif isinstance(folder_list, str):
        logger.warning(f"Converting to list: {folder_list} to a list")
        folder = [folder_list]

    for folder in folder_list:
        folder = os.path.join(base_folder, folder)
        if not os.path.exists(folder):
            os.makedirs(folder)
            logger.info(f"created {folder}")


def isoencode(obj):
    """
    isoencode decodes many different objects such as np.bool -> regular bool.

    ```python
    with open(log_file_path, "a+") as fp:
        json.dump(self.report, fp, default=isoencode)
    ```
    """
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    if isinstance(obj, datetime.date):
        return obj.isoformat()
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, np.int64):
        return int(obj)
    if isinstance(obj, np.float64):
        return float(obj)
    if isinstance(obj, np.bool_):
        return bool(obj)


def save_records(data_inp, output, is_flush=None, write_mode=None):
    """Save list of dicts to file. Instead of loading pandas for such a simple job, this function does the work in most cases.

    :param data_inp: dict or list of dict to be saved
    :param output: path to output file
    :is_flush: whether to flush data to file for each row written to file
    :return: None
    """

    if write_mode is None:
        write_mode = "a+"

    if is_flush is None:
        is_flush = False

    if isinstance(data_inp, list):
        data = data_inp
    elif isinstance(data_inp, dict):
        data = [data_inp]
    else:
        raise Exception("Input data is neither list nor dict: {}".format(data_inp))

    try:
        with open(output, write_mode) as fp:
            for i in data:
                json.dump(i, fp)
                fp.write("\n")
                if is_flush:
                    fp.flush()
    except Exception as ee:
        raise Exception("Could not load data to file: {}".format(ee))


def load_records(data_path_inp):
    """Load data from a line deliminated json file. Instead of loading pandas for such a simple job, this function does the work in most cases.

    :param data_path_inp: data file path
    :return: list of dicts
    """

    data = []

    with open(data_path_inp, "r") as fp:
        for line in fp:
            line = line.replace("null", ' "None" ')
            try:
                line_data = json.loads(line.strip())
            except Exception as ee:
                logger.warning("could not load ", line, "\n", ee)
            data.append(line_data)

    return data



import os

from setuptools import find_packages as _find_packages
from setuptools import setup as _setup

from haferml.version import __version__

# read the contents of your README file
__CWD__ = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(__CWD__, "README.md"), encoding="utf-8") as fp:
    PACKAGE_LONG_DESCRIPTION = fp.read()

PACKAGE_NAME = "haferml"
PACKAGE_VERSION = __version__
PACKAGE_DESCRIPTION = "HAFER ML, the homemade framework for machine learning."
PACKAGE_LONG_DESCRIPTION = PACKAGE_LONG_DESCRIPTION
PACKAGE_URL = "https://github.com/emptymalei/haferml"


def _requirements():
    return [r for r in open("requirements.txt")]


def get_extra_requires(path, add_all=True):
    """
    get_extra_requires retrieves the extras requirements.

    Reference:
    https://hanxiao.io/2019/11/07/A-Better-Practice-for-Managing-extras-require-Dependencies-in-Python/

    :param path: path to the extras require specification
    :type path: str
    :param add_all: whether to include the keyword all, defaults to True
    :type add_all: bool, optional
    :return: The mapping of all the dependencies by extras keyword
    :rtype: dict
    """
    import re
    from collections import defaultdict

    with open(path) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith("#"):
                tags = set()
                if ":" in k:
                    k, v = k.split(":")
                    tags.update(vv.strip() for vv in v.split(","))
                tags.add(re.split("[<=>]", k)[0])
                for t in tags:
                    extra_deps[t].add(k)

        # add tag `all` at the end
        if add_all:
            extra_deps["all"] = set(vv for v in extra_deps.values() for vv in v)

    return extra_deps


def setup():
    _setup(
        name=PACKAGE_NAME,
        version=PACKAGE_VERSION,
        description=PACKAGE_DESCRIPTION,
        long_description=PACKAGE_LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        url=PACKAGE_URL,
        author="L Ma",
        author_email="hi@leima.is",
        license="MIT",
        packages=_find_packages(exclude=("tests",)),
        install_requires=_requirements(),
        include_package_data=True,
        test_suite="nose.collector",
        tests_require=["nose"],
        extras_require=get_extra_requires("requirements.extras.txt"),
        zip_safe=False,
    )


if __name__ == "__main__":
    setup()
    print(_find_packages(exclude=("tests",)))



import inspect
import time
from functools import wraps

from loguru import logger


def attributes(**attrs):
    """
    A decorator to set attributes of member functions in a class.

    ```
    class AGoodClass:
        def __init__(self):
            self.size = 0

        @attributes(order=1)
        def first_good_member(self, new):
            return "first good member"

        @attributes(order=2)
        def second_good_member(self, new):
            return "second good member"
    ```

    References:
    1. https://stackoverflow.com/a/48146924/1477359

    """

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        for attr_name, attr_value in attrs.items():
            setattr(wrapper, attr_name, attr_value)

        return wrapper

    return decorator


class OrderedProcessor:
    """
    Go through an ordered methods in OrderedProcessor to transform a dataframe.

    Beware of the exceptions.

    Add member functions to transform the dataframe

    ```
    @attribute(order=1)
    def _transformer_created_at(self, dataframe):
        pass
    ```
    """

    def __init__(self, **params):

        self.params = params
        self._get_transforms()

        # get current timestamp
        self.current_timestamp = int(time.time())
        logger.info(f"current timestamp: {self.current_timestamp}")

    def _get_transforms(self):
        """
        _get_transforms extracts the list of transformers
        """

        all_methods = dict(inspect.getmembers(self))
        transforms = []

        for method_name, method_func in all_methods.items():
            if hasattr(method_func, "order"):
                method_order = method_func.order
                logger.info(f"{method_name} has order {method_order}")

                transforms.append(
                    {"name": method_name, "method": method_func, "order": method_order}
                )
        transforms = sorted(transforms, key=lambda k: k["order"])
        self.transforms = {m["name"]: m["method"] for m in transforms}

        logger.debug("All methods: {}".format(all_methods))
        logger.info("Ordered predefined transformers: {}".format(self.transforms))



from loguru import logger


class ModelWorkflow:
    """ModelWorkflow class that holds DataSet and ModelSet.

    !!! warning
        This class is a proposed framework. There are many member functions to be implemented if you are using the `ModelWorkflow` class.

        In most projects, we do not need to use this class. We have a class `ModelWorkflowX` which is a more detailed implementation in our `model.pipeline` module.

    `ModelWorkflow` takes a few arguments to instantiate. To run the workflow, use the `train` method.

    :param config: a dictionary that contains the configs.
    :type config: dict
    :param dataset: a DataSet object that contains the data and provides a `create_train_test_datasets` method.
    :type dataset: haferml.model.DataSet
    :param modelset: a ModelSet object that contains the model as well as the hyperparameters and a `create_model`.
    :type modelset: haferml.model.ModelSet
    :param base_folder: working directory where all the artifacts are being perserved.
    :type base_folder: str
    """

    def __init__(self, config, dataset, modelset, base_folder):
        self.config = config
        self.base_folder = base_folder
        self.DataSet = dataset
        self.ModelSet = modelset

    def fit_and_report(self):
        """
        `_fit_and_report` fits the model using input data and generate reports.

        !!! warning
            Please implement this method.
        """
        ...

    def export_results(self):
        """
        export_results saves the necessary artifacts

        !!! warning
            Please implement this method.
        """
        ...

    def train(self, dataset):
        """
        train connects the training workflow and executes the workflow step by step.
        """

        logger.info("1. Create train test datasets")
        self.DataSet.create_train_test_datasets(dataset)
        logger.info("2. Create model")
        self.ModelSet.create_model()
        logger.info("3. Fit model and report")
        self.fit_and_report()
        logger.info("4. Export results")
        self.export_results()
